MODULE transfert_mpi_mod
USE genmod
USE field_mod
  
  TYPE array
    INTEGER,POINTER :: value(:)
    INTEGER,POINTER :: sign(:)
    INTEGER         :: domain
    INTEGER         :: rank
    INTEGER         :: size
    INTEGER,POINTER :: buffer(:)
    REAL,POINTER    :: buffer_r2(:)
    REAL,POINTER    :: buffer_r3(:,:)
    REAL,POINTER    :: buffer_r4(:,:,:)
  END TYPE array
  
  TYPE t_buffer
    REAL,POINTER    :: r2(:)
    REAL,POINTER    :: r3(:,:)
    REAL,POINTER    :: r4(:,:,:)
  END TYPE t_buffer    
    
  TYPE t_request
    INTEGER :: type_field
    INTEGER :: max_size
    INTEGER :: size
    LOGICAL :: vector
    INTEGER,POINTER :: src_domain(:)
    INTEGER,POINTER :: src_i(:)
    INTEGER,POINTER :: src_j(:)
    INTEGER,POINTER :: src_ind(:)
    INTEGER,POINTER :: target_ind(:)
    INTEGER,POINTER :: target_i(:)
    INTEGER,POINTER :: target_j(:)
    INTEGER,POINTER :: target_sign(:)
    INTEGER :: nrecv
    TYPE(ARRAY),POINTER :: recv(:)
    INTEGER :: nsend
    TYPE(ARRAY),POINTER :: send(:)
  END TYPE t_request
  
  TYPE(t_request),POINTER :: req_i1(:)
  TYPE(t_request),POINTER :: req_e1_scal(:)
  TYPE(t_request),POINTER :: req_e1_vect(:)
  
  TYPE(t_request),POINTER :: req_i0(:)
  TYPE(t_request),POINTER :: req_e0_scal(:)
  TYPE(t_request),POINTER :: req_e0_vect(:)
  
  TYPE t_message
    TYPE(t_request), POINTER :: request(:)
    INTEGER :: nreq
    INTEGER, POINTER :: mpi_req(:)
    INTEGER, POINTER :: status(:,:)
    TYPE(t_buffer),POINTER :: buffers(:) 
    TYPE(t_field),POINTER :: field(:)
    LOGICAL :: completed
    LOGICAL :: pending
  END TYPE t_message
  
CONTAINS
  
  SUBROUTINE init_transfert
  USE domain_mod
  USE dimensions
  USE field_mod
  USE metric
  USE mpipara
  IMPLICIT NONE
  INTEGER :: ind,i,j
 
    CALL create_request(field_t,req_i1)

    DO ind=1,ndomain
      CALL swap_dimensions(ind)
      DO i=ii_begin,ii_end+1
        CALL request_add_point(ind,i,jj_begin-1,req_i1)
      ENDDO

      DO j=jj_begin,jj_end
        CALL request_add_point(ind,ii_end+1,j,req_i1)
      ENDDO    
      DO i=ii_begin,ii_end
        CALL request_add_point(ind,i,jj_end+1,req_i1)
      ENDDO    

      DO j=jj_begin,jj_end+1
        CALL request_add_point(ind,ii_begin-1,j,req_i1)
      ENDDO    
    
    ENDDO
  
    CALL finalize_request(req_i1)


    CALL create_request(field_t,req_i0)

    DO ind=1,ndomain
      CALL swap_dimensions(ind)
    
      DO i=ii_begin,ii_end
        CALL request_add_point(ind,i,jj_begin,req_i0)
      ENDDO

      DO j=jj_begin,jj_end
        CALL request_add_point(ind,ii_end,j,req_i0)
      ENDDO    
    
      DO i=ii_begin,ii_end
        CALL request_add_point(ind,i,jj_end,req_i0)
      ENDDO    

      DO j=jj_begin,jj_end
        CALL request_add_point(ind,ii_begin,j,req_i0)
      ENDDO    
    
    ENDDO
 
    CALL finalize_request(req_i0)  


    CALL create_request(field_u,req_e1_scal)
    DO ind=1,ndomain
      CALL swap_dimensions(ind)
      DO i=ii_begin,ii_end
        CALL request_add_point(ind,i,jj_begin-1,req_e1_scal,rup)
        CALL request_add_point(ind,i+1,jj_begin-1,req_e1_scal,lup)
      ENDDO

      DO j=jj_begin,jj_end
        CALL request_add_point(ind,ii_end+1,j,req_e1_scal,left)
        CALL request_add_point(ind,ii_end+1,j-1,req_e1_scal,lup)
      ENDDO    
    
      DO i=ii_begin,ii_end
        CALL request_add_point(ind,i,jj_end+1,req_e1_scal,ldown)
        CALL request_add_point(ind,i-1,jj_end+1,req_e1_scal,rdown)
      ENDDO    

      DO j=jj_begin,jj_end
        CALL request_add_point(ind,ii_begin-1,j,req_e1_scal,right)
        CALL request_add_point(ind,ii_begin-1,j+1,req_e1_scal,rdown)
      ENDDO   

    ENDDO

    CALL finalize_request(req_e1_scal)


    CALL create_request(field_u,req_e0_scal)
    DO ind=1,ndomain
      CALL swap_dimensions(ind)


      DO i=ii_begin+1,ii_end-1
        CALL request_add_point(ind,i,jj_begin,req_e0_scal,right)
        CALL request_add_point(ind,i,jj_end,req_e0_scal,right)
      ENDDO
    
      DO j=jj_begin+1,jj_end-1
        CALL request_add_point(ind,ii_begin,j,req_e0_scal,rup)
        CALL request_add_point(ind,ii_end,j,req_e0_scal,rup)
      ENDDO   

      CALL request_add_point(ind,ii_begin+1,jj_begin,req_e0_scal,left)
      CALL request_add_point(ind,ii_begin,jj_begin+1,req_e0_scal,ldown)
      CALL request_add_point(ind,ii_begin+1,jj_end,req_e0_scal,left)
      CALL request_add_point(ind,ii_end,jj_begin+1,req_e0_scal,ldown)

    ENDDO

    CALL finalize_request(req_e0_scal)


    
    CALL create_request(field_u,req_e1_vect,.TRUE.)
    DO ind=1,ndomain
      CALL swap_dimensions(ind)
      DO i=ii_begin,ii_end
        CALL request_add_point(ind,i,jj_begin-1,req_e1_vect,rup)
        CALL request_add_point(ind,i+1,jj_begin-1,req_e1_vect,lup)
      ENDDO

      DO j=jj_begin,jj_end
        CALL request_add_point(ind,ii_end+1,j,req_e1_vect,left)
        CALL request_add_point(ind,ii_end+1,j-1,req_e1_vect,lup)
      ENDDO    
    
      DO i=ii_begin,ii_end
        CALL request_add_point(ind,i,jj_end+1,req_e1_vect,ldown)
        CALL request_add_point(ind,i-1,jj_end+1,req_e1_vect,rdown)
      ENDDO    

      DO j=jj_begin,jj_end
        CALL request_add_point(ind,ii_begin-1,j,req_e1_vect,right)
        CALL request_add_point(ind,ii_begin-1,j+1,req_e1_vect,rdown)
      ENDDO   

  
    ENDDO  

    CALL finalize_request(req_e1_vect)
    
    
    CALL create_request(field_u,req_e0_vect,.TRUE.)
    DO ind=1,ndomain
      CALL swap_dimensions(ind)
 
      DO i=ii_begin+1,ii_end-1
        CALL request_add_point(ind,i,jj_begin,req_e0_vect,right)
        CALL request_add_point(ind,i,jj_end,req_e0_vect,right)
      ENDDO
    
      DO j=jj_begin+1,jj_end-1
        CALL request_add_point(ind,ii_begin,j,req_e0_vect,rup)
        CALL request_add_point(ind,ii_end,j,req_e0_vect,rup)
      ENDDO   

      CALL request_add_point(ind,ii_begin+1,jj_begin,req_e0_vect,left)
      CALL request_add_point(ind,ii_begin,jj_begin+1,req_e0_vect,ldown)
      CALL request_add_point(ind,ii_begin+1,jj_end,req_e0_vect,left)
      CALL request_add_point(ind,ii_end,jj_begin+1,req_e0_vect,ldown)
  
    ENDDO  

    CALL finalize_request(req_e0_vect)


  END SUBROUTINE init_transfert
  
  SUBROUTINE create_request(type_field,request,vector)
  USE domain_mod
  USE field_mod
  IMPLICIT NONE
    INTEGER :: type_field
    TYPE(t_request),POINTER :: request(:)
    LOGICAL,OPTIONAL :: vector
    
    TYPE(t_request),POINTER :: req
    TYPE(t_domain),POINTER :: d
    INTEGER :: ind
    INTEGER :: max_size
       
    ALLOCATE(request(ndomain))

    DO ind=1,ndomain
      req=>request(ind)
      d=>domain(ind)
      IF (type_field==field_t) THEN
        Max_size=2*(d%iim+2)+2*(d%jjm+2)
      ELSE IF (type_field==field_u) THEN
        Max_size=3*(2*(d%iim+2)+2*(d%jjm+2))
      ELSE IF (type_field==field_z) THEN
        Max_size=2*(2*(d%iim+2)+2*(d%jjm+2))
      ENDIF

      req%type_field=type_field
      req%max_size=max_size*2
      req%size=0
      req%vector=.FALSE.
      IF (PRESENT(vector)) req%vector=vector
      ALLOCATE(req%src_domain(req%max_size))
      ALLOCATE(req%src_ind(req%max_size))
      ALLOCATE(req%target_ind(req%max_size))
      ALLOCATE(req%src_i(req%max_size))
      ALLOCATE(req%src_j(req%max_size))
      ALLOCATE(req%target_i(req%max_size))
      ALLOCATE(req%target_j(req%max_size))
      ALLOCATE(req%target_sign(req%max_size))
    ENDDO
  
  END SUBROUTINE create_request

  SUBROUTINE reallocate_request(req)
  IMPLICIT NONE
    TYPE(t_request),POINTER :: req
      
    INTEGER,POINTER :: src_domain(:)
    INTEGER,POINTER :: src_ind(:)
    INTEGER,POINTER :: target_ind(:)
    INTEGER,POINTER :: src_i(:)
    INTEGER,POINTER :: src_j(:)
    INTEGER,POINTER :: target_i(:)
    INTEGER,POINTER :: target_j(:)
    INTEGER,POINTER :: target_sign(:)

    PRINT *,"REALLOCATE_REQUEST"
    src_domain=>req%src_domain
    src_ind=>req%src_ind
    target_ind=>req%target_ind
    src_i=>req%src_i
    src_j=>req%src_j
    target_i=>req%target_i
    target_j=>req%target_j
    target_sign=>req%target_sign

    ALLOCATE(req%src_domain(req%max_size*2))
    ALLOCATE(req%src_ind(req%max_size*2))
    ALLOCATE(req%target_ind(req%max_size*2))
    ALLOCATE(req%src_i(req%max_size*2))
    ALLOCATE(req%src_j(req%max_size*2))
    ALLOCATE(req%target_i(req%max_size*2))
    ALLOCATE(req%target_j(req%max_size*2))
    ALLOCATE(req%target_sign(req%max_size*2))
    
    req%src_domain(1:req%max_size)=src_domain(:)
    req%src_ind(1:req%max_size)=src_ind(:)
    req%target_ind(1:req%max_size)=target_ind(:)
    req%src_i(1:req%max_size)=src_i(:)
    req%src_j(1:req%max_size)=src_j(:)
    req%target_i(1:req%max_size)=target_i(:)
    req%target_j(1:req%max_size)=target_j(:)
    req%target_sign(1:req%max_size)=target_sign(:)
    
    req%max_size=req%max_size*2
         
    DEALLOCATE(src_domain)
    DEALLOCATE(src_ind)
    DEALLOCATE(target_ind)
    DEALLOCATE(src_i)
    DEALLOCATE(src_j)
    DEALLOCATE(target_i)
    DEALLOCATE(target_j)
    DEALLOCATE(target_sign)

  END SUBROUTINE reallocate_request

      
    SUBROUTINE request_add_point(ind,i,j,request,pos)
    USE domain_mod
    USE field_mod
    IMPLICIT NONE
      INTEGER,INTENT(IN)            ::  ind
      INTEGER,INTENT(IN)            :: i
      INTEGER,INTENT(IN)            :: j
      TYPE(t_request),POINTER :: request(:)
      INTEGER,INTENT(IN),OPTIONAL  :: pos
      
      INTEGER :: src_domain
      INTEGER :: src_iim,src_i,src_j,src_n,src_pos,src_delta
      TYPE(t_request),POINTER :: req
      TYPE(t_domain),POINTER :: d
      
      req=>request(ind)
      d=>domain(ind)
      
      IF (req%max_size==req%size) CALL reallocate_request(req)
      req%size=req%size+1
      IF (req%type_field==field_t) THEN
        src_domain=domain(ind)%assign_domain(i,j)
        src_iim=domain_glo(src_domain)%iim
        src_i=domain(ind)%assign_i(i,j)
        src_j=domain(ind)%assign_j(i,j)

        req%target_ind(req%size)=(j-1)*d%iim+i
        req%target_sign(req%size)=1
        req%src_domain(req%size)=src_domain
        req%src_ind(req%size)=(src_j-1)*src_iim+src_i
      ELSE IF (req%type_field==field_u) THEN
        IF (.NOT. PRESENT(pos)) STOP 'argument request_add_point non conforme'

        src_domain=domain(ind)%edge_assign_domain(pos-1,i,j)
        src_iim=domain_glo(src_domain)%iim
        src_i=domain(ind)%edge_assign_i(pos-1,i,j)
        src_j=domain(ind)%edge_assign_j(pos-1,i,j)
        src_n=(src_j-1)*src_iim+src_i
        src_delta=domain(ind)%delta(i,j)
        src_pos=domain(ind)%edge_assign_pos(pos-1,i,j)+1
                
        req%target_ind(req%size)=(j-1)*d%iim+i+d%u_pos(pos)

        req%target_sign(req%size)= 1
        IF (req%vector) req%target_sign(req%size)= domain(ind)%edge_assign_sign(pos-1,i,j)

        req%src_domain(req%size)=src_domain
        req%src_ind(req%size)=src_n+domain_glo(src_domain)%u_pos(src_pos)

      ELSE IF (req%type_field==field_z) THEN
        IF (.NOT. PRESENT(pos)) STOP 'argument request_add_point non conforme'

        src_domain=domain(ind)%assign_domain(i,j)
        src_iim=domain_glo(src_domain)%iim
        src_i=domain(ind)%assign_i(i,j)
        src_j=domain(ind)%assign_j(i,j)
        src_n=(src_j-1)*src_iim+src_i
        src_delta=domain(ind)%delta(i,j)
        
        src_pos=MOD(pos-1+src_delta+6,6)+1
        
        req%target_ind(req%size)=(j-1)*d%iim+i+d%z_pos(pos)
        req%target_sign(req%size)=1
        req%src_domain(req%size)=src_domain
        req%src_ind(req%size)=src_n+domain_glo(src_domain)%z_pos(src_pos)
      ENDIF
  END SUBROUTINE request_add_point
  
  
  SUBROUTINE Finalize_request(request)
  USE mpipara
  USE domain_mod
  USE mpi_mod
  IMPLICIT NONE
    TYPE(t_request),POINTER :: request(:)
    TYPE(t_request),POINTER :: req
    INTEGER :: nb_domain_recv(0:mpi_size-1)
    INTEGER :: nb_domain_send(0:mpi_size-1)
    INTEGER :: nb_data_domain_recv(ndomain_glo)
    INTEGER :: list_domain_recv(ndomain_glo)
    INTEGER,ALLOCATABLE :: list_domain_send(:)
    INTEGER             :: list_domain(ndomain)

    INTEGER :: rank,i,j
    INTEGER :: size,ind_glo,ind_loc
    INTEGER :: isend, irecv, ireq, nreq
    INTEGER, ALLOCATABLE :: mpi_req(:)
    INTEGER, ALLOCATABLE :: status(:,:)
    
    IF (.NOT. using_mpi) RETURN
    
    DO ind_loc=1,ndomain
      req=>request(ind_loc)
      
      nb_data_domain_recv(:) = 0
      nb_domain_recv(:) = 0
      
      DO i=1,req%size
        ind_glo=req%src_domain(i)
        nb_data_domain_recv(ind_glo)=nb_data_domain_recv(ind_glo)+1
      ENDDO
 
      DO ind_glo=1,ndomain_glo
        IF ( nb_data_domain_recv(ind_glo) > 0 )  nb_domain_recv(domglo_rank(ind_glo))=nb_domain_recv(domglo_rank(ind_glo))+1
      ENDDO

      req%nrecv=sum(nb_domain_recv(:))
      ALLOCATE(req%recv(req%nrecv))

      irecv=0
      DO ind_glo=1,ndomain_glo
        IF (nb_data_domain_recv(ind_glo)>0) THEN
          irecv=irecv+1
          list_domain_recv(ind_glo)=irecv
          req%recv(irecv)%rank=domglo_rank(ind_glo)
          req%recv(irecv)%size=nb_data_domain_recv(ind_glo)
          req%recv(irecv)%domain=domglo_loc_ind(ind_glo)
          ALLOCATE(req%recv(irecv)%value(req%recv(irecv)%size))
          ALLOCATE(req%recv(irecv)%sign(req%recv(irecv)%size))
          ALLOCATE(req%recv(irecv)%buffer(req%recv(irecv)%size))
        ENDIF
      ENDDO
      
      req%recv(:)%size=0
      irecv=0
      DO i=1,req%size
        irecv=list_domain_recv(req%src_domain(i))
        req%recv(irecv)%size=req%recv(irecv)%size+1
        size=req%recv(irecv)%size
        req%recv(irecv)%value(size)=req%src_ind(i)
        req%recv(irecv)%buffer(size)=req%target_ind(i)
        req%recv(irecv)%sign(size)=req%target_sign(i)
      ENDDO
    ENDDO

    nb_domain_recv(:) = 0    
    DO ind_loc=1,ndomain
      req=>request(ind_loc)
      
      DO irecv=1,req%nrecv
        rank=req%recv(irecv)%rank
        nb_domain_recv(rank)=nb_domain_recv(rank)+1
      ENDDO
    ENDDO
    
    CALL MPI_Alltoall(nb_domain_recv,1,MPI_INTEGER,nb_domain_send,1,MPI_INTEGER,comm_icosa,ierr)     
    

    ALLOCATE(list_domain_send(sum(nb_domain_send)))
    
    nreq=sum(nb_domain_recv(:))+sum(nb_domain_send(:))
    ALLOCATE(mpi_req(nreq))
    ALLOCATE(status(MPI_STATUS_SIZE,nreq))
    
    ireq=0
    DO ind_loc=1,ndomain
      req=>request(ind_loc)
      DO irecv=1,req%nrecv
        ireq=ireq+1
        CALL MPI_ISEND(req%recv(irecv)%domain,1,MPI_INTEGER,req%recv(irecv)%rank,0,comm_icosa, mpi_req(ireq),ierr)
      ENDDO
    ENDDO
    
    j=0
    DO rank=0,mpi_size-1
      DO i=1,nb_domain_send(rank)
        j=j+1
        ireq=ireq+1
        CALL MPI_IRECV(list_domain_send(j),1,MPI_INTEGER,rank,0,comm_icosa, mpi_req(ireq),ierr)
      ENDDO
    ENDDO
    
    CALL MPI_WAITALL(nreq,mpi_req,status,ierr)
    
    list_domain(:)=0
    DO i=1,sum(nb_domain_send)
      ind_loc=list_domain_send(i)
      list_domain(ind_loc)=list_domain(ind_loc)+1
    ENDDO
    
    DO ind_loc=1,ndomain
      req=>request(ind_loc)
      req%nsend=list_domain(ind_loc)
      ALLOCATE(req%send(req%nsend))
    ENDDO
   
   ireq=0 
   DO ind_loc=1,ndomain
     req=>request(ind_loc)
     
     DO irecv=1,req%nrecv
       ireq=ireq+1
       CALL MPI_ISEND(mpi_rank,1,MPI_INTEGER,req%recv(irecv)%rank,req%recv(irecv)%domain,comm_icosa, mpi_req(ireq),ierr)
     ENDDO
     
     DO isend=1,req%nsend
       ireq=ireq+1
       CALL MPI_IRECV(req%send(isend)%rank,1,MPI_INTEGER,MPI_ANY_SOURCE,ind_loc,comm_icosa, mpi_req(ireq),ierr)
     ENDDO
   ENDDO

   CALL MPI_WAITALL(nreq,mpi_req,status,ierr)
   CALL MPI_BARRIER(comm_icosa,ierr)

   ireq=0 
   DO ind_loc=1,ndomain
     req=>request(ind_loc)
     
     DO irecv=1,req%nrecv
       ireq=ireq+1
       CALL MPI_ISEND(req%recv(irecv)%size,1,MPI_INTEGER,req%recv(irecv)%rank,req%recv(irecv)%domain,comm_icosa, mpi_req(ireq),ierr)
     ENDDO
     
     DO isend=1,req%nsend
       ireq=ireq+1
       CALL MPI_IRECV(req%send(isend)%size,1,MPI_INTEGER,req%send(isend)%rank,ind_loc,comm_icosa, mpi_req(ireq),ierr)
     ENDDO
   ENDDO

   CALL MPI_WAITALL(nreq,mpi_req,status,ierr)

   ireq=0 
   DO ind_loc=1,ndomain
     req=>request(ind_loc)
     
     DO irecv=1,req%nrecv
       ireq=ireq+1
       CALL MPI_ISEND(req%recv(irecv)%value,req%recv(irecv)%size,MPI_INTEGER,&
            req%recv(irecv)%rank,req%recv(irecv)%domain,comm_icosa, mpi_req(ireq),ierr)
     ENDDO
     
     DO isend=1,req%nsend
       ireq=ireq+1
       ALLOCATE(req%send(isend)%value(req%send(isend)%size))
       CALL MPI_IRECV(req%send(isend)%value,req%send(isend)%size,MPI_INTEGER,&
            req%send(isend)%rank,ind_loc,comm_icosa, mpi_req(ireq),ierr)
     ENDDO
   ENDDO

   CALL MPI_WAITALL(nreq,mpi_req,status,ierr)

   DO ind_loc=1,ndomain
     req=>request(ind_loc)
     
     DO irecv=1,req%nrecv
       req%recv(irecv)%value(:)=req%recv(irecv)%buffer(:)
       req%recv(irecv)%sign(:) =req%recv(irecv)%sign(:)
       DEALLOCATE(req%recv(irecv)%buffer)
     ENDDO
   ENDDO  
   
    
  END SUBROUTINE Finalize_request 


  SUBROUTINE init_message_seq(field, request, message)
  USE field_mod
  USE domain_mod
  USE mpi_mod
  USE mpipara
  USE mpi_mod
  IMPLICIT NONE
    TYPE(t_field),POINTER :: field(:)
    TYPE(t_request),POINTER :: request(:)
    TYPE(t_message) :: message

!$OMP MASTER    
    message%request=>request
!$OMP END MASTER    
!$OMP BARRIER    

  END SUBROUTINE init_message_seq

  SUBROUTINE send_message_seq(field,message)
  USE field_mod
  USE domain_mod
  USE mpi_mod
  USE mpipara
  USE omp_para
  USE trace
  IMPLICIT NONE
    TYPE(t_field),POINTER :: field(:)
    TYPE(t_message) :: message

    CALL transfert_request_seq(field,message%request)
    
  END SUBROUTINE send_message_seq
  
  SUBROUTINE test_message_seq(message)
  IMPLICIT NONE
    TYPE(t_message) :: message
  END SUBROUTINE  test_message_seq
  
   
  SUBROUTINE wait_message_seq(message)
  IMPLICIT NONE
    TYPE(t_message) :: message
    
  END SUBROUTINE wait_message_seq    

  SUBROUTINE transfert_message_seq(field,message)
  USE field_mod
  USE domain_mod
  USE mpi_mod
  USE mpipara
  USE omp_para
  USE trace
  IMPLICIT NONE
    TYPE(t_field),POINTER :: field(:)
    TYPE(t_message) :: message

   CALL send_message_seq(field,message)
    
  END SUBROUTINE transfert_message_seq    
    
    
  SUBROUTINE init_message_mpi(field,request, message)
  USE field_mod
  USE domain_mod
  USE mpi_mod
  USE mpipara
  USE mpi_mod
  IMPLICIT NONE
  
    TYPE(t_field),POINTER :: field(:)
    TYPE(t_request),POINTER :: request(:)
    TYPE(t_message) :: message

    TYPE(ARRAY),POINTER :: recv,send 
    TYPE(t_request),POINTER :: req
    INTEGER :: irecv,isend
    INTEGER :: ireq,nreq
    INTEGER :: ind
    INTEGER :: dim3,dim4

!$OMP MASTER
    message%request=>request
    nreq=sum(request(:)%nsend)+sum(request(:)%nrecv)
    message%nreq=nreq
    ALLOCATE(message%mpi_req(nreq))
    ALLOCATE(message%buffers(nreq))
    ALLOCATE(message%status(MPI_STATUS_SIZE,nreq))
    
    message%pending=.FALSE.
    message%completed=.FALSE.
    
    IF (field(1)%data_type==type_real) THEN

      IF (field(1)%ndim==2) THEN
      
        ireq=0
        DO ind=1,ndomain
          req=>request(ind)
      
          DO isend=1,req%nsend
            ireq=ireq+1
            send=>req%send(isend)
            CALL allocate_mpi_buffer(message%buffers(ireq)%r2,send%size)
          ENDDO
        
          DO irecv=1,req%nrecv
            ireq=ireq+1
            recv=>req%recv(irecv)
            CALL allocate_mpi_buffer(message%buffers(ireq)%r2,recv%size)
          ENDDO
        
        ENDDO
      
      
      ELSE  IF (field(1)%ndim==3) THEN
    
        ireq=0
        DO ind=1,ndomain
          dim3=size(field(ind)%rval3d,2)
          req=>request(ind)
 
          DO isend=1,req%nsend
            ireq=ireq+1
            send=>req%send(isend)
            CALL allocate_mpi_buffer(message%buffers(ireq)%r3,send%size,dim3)
          ENDDO
        
          DO irecv=1,req%nrecv
            ireq=ireq+1
            recv=>req%recv(irecv)
            CALL allocate_mpi_buffer(message%buffers(ireq)%r3,recv%size,dim3)

          ENDDO
        
        ENDDO


      ELSE  IF (field(1)%ndim==4) THEN
   
        ireq=0
        DO ind=1,ndomain
          dim3=size(field(ind)%rval4d,2)
          dim4=size(field(ind)%rval4d,3)
          req=>request(ind)

          DO isend=1,req%nsend
            ireq=ireq+1
            send=>req%send(isend)
            CALL allocate_mpi_buffer(message%buffers(ireq)%r4,send%size,dim3,dim4)
          ENDDO
        
          DO irecv=1,req%nrecv
            ireq=ireq+1
            recv=>req%recv(irecv)
            CALL allocate_mpi_buffer(message%buffers(ireq)%r4,recv%size,dim3,dim4)
          ENDDO
        
        ENDDO
      
      ENDIF      
    ENDIF
!$OMP END MASTER
!$OMP BARRIER    
  END SUBROUTINE init_message_mpi
  
  SUBROUTINE barrier
  USE mpi_mod
  USE mpipara
  IMPLICIT NONE
    
    CALL MPI_BARRIER(comm_icosa,ierr)
    
  END SUBROUTINE barrier  
    
  SUBROUTINE transfert_message_mpi(field,message)
  USE field_mod
  IMPLICIT NONE
    TYPE(t_field),POINTER :: field(:)
    TYPE(t_message) :: message
    
    CALL send_message_mpi(field,message)
    CALL wait_message_mpi(message)
    
  END SUBROUTINE transfert_message_mpi


  SUBROUTINE send_message_mpi(field,message)
  USE field_mod
  USE domain_mod
  USE mpi_mod
  USE mpipara
  USE omp_para
  USE trace
  IMPLICIT NONE
    TYPE(t_field),POINTER :: field(:)
    TYPE(t_message) :: message
    REAL(rstd),POINTER :: rval2d(:) 
    REAL(rstd),POINTER :: rval3d(:,:) 
    REAL(rstd),POINTER :: rval4d(:,:,:) 
    REAL(rstd),POINTER :: buffer_r2(:) 
    REAL(rstd),POINTER :: buffer_r3(:,:) 
    REAL(rstd),POINTER :: buffer_r4(:,:,:) 
    INTEGER,POINTER :: value(:) 
    INTEGER,POINTER :: sgn(:) 
    TYPE(ARRAY),POINTER :: recv,send 
    TYPE(t_request),POINTER :: req
    INTEGER, ALLOCATABLE :: mpi_req(:)
    INTEGER, ALLOCATABLE :: status(:,:)
    INTEGER :: irecv,isend
    INTEGER :: ireq,nreq
    INTEGER :: ind,n,l,m
    INTEGER :: dim3,dim4

!$OMP BARRIER

    CALL trace_start("transfert_mpi")

    nreq=message%nreq
    message%field=>field

!$OMP MASTER
    message%completed=.FALSE.
    message%pending=.TRUE.
!$OMP END MASTER
    
    IF (field(1)%data_type==type_real) THEN
      IF (field(1)%ndim==2) THEN

        ireq=0
        DO ind=1,ndomain
          rval2d=>field(ind)%rval2d
        
          req=>message%request(ind)
          DO isend=1,req%nsend
            ireq=ireq+1
            send=>req%send(isend)
            buffer_r2=>message%buffers(ireq)%r2
            value=>send%value

            CALL trace_in

!$OMP DO SCHEDULE(STATIC)
            DO n=1,send%size
              buffer_r2(n)=rval2d(value(n))
            ENDDO
            
            CALL trace_out

!$OMP MASTER
            CALL MPI_ISSEND(buffer_r2,send%size,MPI_REAL8,send%rank,ind,comm_icosa, message%mpi_req(ireq),ierr)
!$OMP END MASTER
          ENDDO
        
          DO irecv=1,req%nrecv
            ireq=ireq+1
            recv=>req%recv(irecv)
            buffer_r2=>message%buffers(ireq)%r2
!$OMP MASTER
            CALL MPI_IRECV(buffer_r2,recv%size,MPI_REAL8,recv%rank,recv%domain,comm_icosa, message%mpi_req(ireq),ierr)
!$OMP END MASTER
          ENDDO
        
        ENDDO
      
      ELSE  IF (field(1)%ndim==3) THEN
      
        ireq=0
        DO ind=1,ndomain
          dim3=size(field(ind)%rval3d,2)
          rval3d=>field(ind)%rval3d
          req=>message%request(ind)
 
          DO isend=1,req%nsend
            ireq=ireq+1
            send=>req%send(isend)
            buffer_r3=>message%buffers(ireq)%r3
            value=>send%value

            CALL trace_in
            
!$OMP DO SCHEDULE(STATIC)
              DO n=1,send%size
                buffer_r3(n,:)=rval3d(value(n),:)
              ENDDO

             CALL trace_out

!$OMP MASTER
            CALL MPI_ISSEND(buffer_r3,send%size*dim3,MPI_REAL8,send%rank,ind,comm_icosa, message%mpi_req(ireq),ierr)
!$OMP END MASTER
          ENDDO
        
          DO irecv=1,req%nrecv
            ireq=ireq+1
            recv=>req%recv(irecv)
            buffer_r3=>message%buffers(ireq)%r3
!$OMP MASTER           
            CALL MPI_IRECV(buffer_r3,recv%size*dim3,MPI_REAL8,recv%rank,recv%domain,comm_icosa, message%mpi_req(ireq),ierr)
!$OMP END MASTER

          ENDDO
        
        ENDDO

      ELSE  IF (field(1)%ndim==4) THEN
    
        ireq=0
        DO ind=1,ndomain
          dim3=size(field(ind)%rval4d,2)
          dim4=size(field(ind)%rval4d,3)
          rval4d=>field(ind)%rval4d
          req=>message%request(ind)

          DO isend=1,req%nsend
            ireq=ireq+1
            send=>req%send(isend)
            buffer_r4=>message%buffers(ireq)%r4
            value=>send%value

            CALL trace_in

!$OMP DO SCHEDULE(STATIC)
            DO n=1,send%size
               buffer_r4(n,:,:)=rval4d(value(n),:,:)
            ENDDO

           CALL trace_out

!$OMP MASTER
            CALL MPI_ISSEND(buffer_r4,send%size*dim3*dim4,MPI_REAL8,send%rank,ind,comm_icosa, message%mpi_req(ireq),ierr)
!$OMP END MASTER
          ENDDO
        
          DO irecv=1,req%nrecv
            ireq=ireq+1
            recv=>req%recv(irecv)
            buffer_r4=>message%buffers(ireq)%r4
!$OMP MASTER           
            CALL MPI_IRECV(buffer_r4,recv%size*dim3*dim4,MPI_REAL8,recv%rank,recv%domain,comm_icosa, message%mpi_req(ireq),ierr)
!$OMP END MASTER
          ENDDO
        
        ENDDO
      
      ENDIF      
      
    ENDIF

    CALL trace_end("transfert_mpi")
!$OMP BARRIER
    
  END SUBROUTINE send_message_mpi
  
  SUBROUTINE test_message_mpi(message)
  IMPLICIT NONE
    TYPE(t_message) :: message
    
    INTEGER :: ierr
!$OMP MASTER
     IF (.NOT. message%pending) RETURN
!$OMP END MASTER

!$OMP MASTER
     IF (.NOT. message%completed) CALL MPI_TESTALL(message%nreq,message%mpi_req,message%completed,message%status,ierr)
!$OMP END MASTER
  END SUBROUTINE  test_message_mpi
  
   
  SUBROUTINE wait_message_mpi(message)
  USE field_mod
  USE domain_mod
  USE mpi_mod
  USE mpipara
  USE omp_para
  USE trace
  IMPLICIT NONE
    TYPE(t_message) :: message

    TYPE(t_field),POINTER :: field(:)
    REAL(rstd),POINTER :: rval2d(:) 
    REAL(rstd),POINTER :: rval3d(:,:) 
    REAL(rstd),POINTER :: rval4d(:,:,:) 
    REAL(rstd),POINTER :: buffer_r2(:) 
    REAL(rstd),POINTER :: buffer_r3(:,:) 
    REAL(rstd),POINTER :: buffer_r4(:,:,:) 
    INTEGER,POINTER :: value(:) 
    INTEGER,POINTER :: sgn(:) 
    TYPE(ARRAY),POINTER :: recv,send 
    TYPE(t_request),POINTER :: req
    INTEGER, ALLOCATABLE :: mpi_req(:)
    INTEGER, ALLOCATABLE :: status(:,:)
    INTEGER :: irecv,isend
    INTEGER :: ireq,nreq
    INTEGER :: ind,n,l,m
    INTEGER :: dim3,dim4

!$OMP BARRIER

    CALL trace_start("transfert_mpi")

    IF (.NOT. message%pending) RETURN
    
    field=>message%field
    nreq=message%nreq
    
    IF (field(1)%data_type==type_real) THEN
      IF (field(1)%ndim==2) THEN

!$OMP MASTER
        IF (.NOT. message%completed) CALL MPI_WAITALL(nreq,message%mpi_req,message%status,ierr)
!$OMP END MASTER
!$OMP BARRIER

        ireq=0        
        DO ind=1,ndomain
          rval2d=>field(ind)%rval2d
          req=>message%request(ind)

          DO isend=1,req%nsend
            ireq=ireq+1
          ENDDO
     
          DO irecv=1,req%nrecv
            ireq=ireq+1
            recv=>req%recv(irecv)
            buffer_r2=>message%buffers(ireq)%r2
            value=>recv%value
            sgn=>recv%sign

            CALL trace_in
            
!$OMP DO SCHEDULE(STATIC)
            DO n=1,recv%size
              rval2d(value(n))=buffer_r2(n)*sgn(n)  
            ENDDO        

            CALL trace_out

          ENDDO
        
        ENDDO
      
      
      ELSE  IF (field(1)%ndim==3) THEN

!$OMP MASTER
        IF (.NOT. message%completed) CALL MPI_WAITALL(nreq,message%mpi_req,message%status,ierr)
!$OMP END MASTER
!$OMP BARRIER

        ireq=0        
        DO ind=1,ndomain
          rval3d=>field(ind)%rval3d
          req=>message%request(ind)

          DO isend=1,req%nsend
            ireq=ireq+1
          ENDDO
        
          DO irecv=1,req%nrecv
            ireq=ireq+1
            recv=>req%recv(irecv)
            buffer_r3=>message%buffers(ireq)%r3
            value=>recv%value
            sgn=>recv%sign

            CALL trace_in
            
!$OMP DO SCHEDULE(STATIC)
            DO n=1,recv%size
              rval3d(value(n),:)=buffer_r3(n,:)*sgn(n)  
            ENDDO  

            CALL trace_out

          ENDDO
        
        ENDDO

      ELSE  IF (field(1)%ndim==4) THEN
!$OMP MASTER
        IF (.NOT. message%completed) CALL MPI_WAITALL(nreq,message%mpi_req,message%status,ierr)
!$OMP END MASTER
!$OMP BARRIER

        ireq=0        
        DO ind=1,ndomain
          rval4d=>field(ind)%rval4d
          req=>message%request(ind)

          DO isend=1,req%nsend
            ireq=ireq+1
          ENDDO
        
          DO irecv=1,req%nrecv
            ireq=ireq+1
            recv=>req%recv(irecv)
            buffer_r4=>message%buffers(ireq)%r4
            value=>recv%value
            sgn=>recv%sign

            CALL trace_in

!$OMP DO SCHEDULE(STATIC)
            DO n=1,recv%size
              rval4d(value(n),:,:)=buffer_r4(n,:,:)*sgn(n) 
            ENDDO

            CALL trace_out

          ENDDO
        
        ENDDO
      
      ENDIF      
      
    ENDIF

!$OMP MASTER
    message%pending=.FALSE.
!$OMP END MASTER

    CALL trace_end("transfert_mpi")
!$OMP BARRIER
    
  END SUBROUTINE wait_message_mpi


  SUBROUTINE transfert_request_mpi(field,request)
  USE field_mod
  USE domain_mod
  USE mpi_mod
  USE mpipara
  USE trace
  IMPLICIT NONE
    TYPE(t_field),POINTER :: field(:)
    TYPE(t_request),POINTER :: request(:)
    REAL(rstd),POINTER :: rval2d(:) 
    REAL(rstd),POINTER :: rval3d(:,:) 
    REAL(rstd),POINTER :: rval4d(:,:,:) 
    REAL(rstd),POINTER :: buffer_r2(:) 
    REAL(rstd),POINTER :: buffer_r3(:,:) 
    REAL(rstd),POINTER :: buffer_r4(:,:,:) 
    INTEGER,POINTER :: value(:) 
    INTEGER,POINTER :: sgn(:) 
    TYPE(ARRAY),POINTER :: recv,send 
    TYPE(t_request),POINTER :: req
    INTEGER, ALLOCATABLE :: mpi_req(:)
    INTEGER, ALLOCATABLE :: status(:,:)
    INTEGER :: irecv,isend
    INTEGER :: ireq,nreq
    INTEGER :: ind,n
    INTEGER :: dim3,dim4

    CALL trace_start("transfert_mpi")

    IF (field(1)%data_type==type_real) THEN
      IF (field(1)%ndim==2) THEN
      
        nreq=sum(request(:)%nsend)+sum(request(:)%nrecv)
        ALLOCATE(mpi_req(nreq))
        ALLOCATE(status(MPI_STATUS_SIZE,nreq))
    
        ireq=0
        DO ind=1,ndomain
          rval2d=>field(ind)%rval2d
        
          req=>request(ind)
          DO isend=1,req%nsend
            send=>req%send(isend)

            ALLOCATE(send%buffer_r2(send%size))
            buffer_r2=>send%buffer_r2
            value=>send%value
            DO n=1,send%size
              buffer_r2(n)=rval2d(value(n))
            ENDDO

            ireq=ireq+1
            CALL MPI_ISEND(buffer_r2,send%size,MPI_REAL8,send%rank,ind,comm_icosa, mpi_req(ireq),ierr)
          ENDDO
        
          DO irecv=1,req%nrecv
            recv=>req%recv(irecv)
            ALLOCATE(recv%buffer_r2(recv%size))
           
            ireq=ireq+1
            CALL MPI_IRECV(recv%buffer_r2,recv%size,MPI_REAL8,recv%rank,recv%domain,comm_icosa, mpi_req(ireq),ierr)
          ENDDO
        
        ENDDO
        
        CALL MPI_WAITALL(nreq,mpi_req,status,ierr)
        
        DO ind=1,ndomain
          rval2d=>field(ind)%rval2d
        
          req=>request(ind)
          DO isend=1,req%nsend
            send=>req%send(isend)
            DEALLOCATE(send%buffer_r2)
          ENDDO
        
          DO irecv=1,req%nrecv
            recv=>req%recv(irecv)
            buffer_r2=>recv%buffer_r2
            value=>recv%value
            sgn=>recv%sign
            DO n=1,recv%size
              rval2d(value(n))=buffer_r2(n)*sgn(n)  
            ENDDO        
            DEALLOCATE(recv%buffer_r2)
          ENDDO
        
        ENDDO
      
      
      ELSE  IF (field(1)%ndim==3) THEN
      
        nreq=sum(request(:)%nsend)+sum(request(:)%nrecv)
        ALLOCATE(mpi_req(nreq))
        ALLOCATE(status(MPI_STATUS_SIZE,nreq))
    
        ireq=0
        DO ind=1,ndomain
          dim3=size(field(ind)%rval3d,2)
          rval3d=>field(ind)%rval3d
        
          req=>request(ind)
          DO isend=1,req%nsend
            send=>req%send(isend)

            ALLOCATE(send%buffer_r3(send%size,dim3))
            buffer_r3=>send%buffer_r3
            value=>send%value
            DO n=1,send%size
              buffer_r3(n,:)=rval3d(value(n),:)
            ENDDO

            ireq=ireq+1
            CALL MPI_ISEND(buffer_r3,send%size*dim3,MPI_REAL8,send%rank,ind,comm_icosa, mpi_req(ireq),ierr)
          ENDDO
        
          DO irecv=1,req%nrecv
            recv=>req%recv(irecv)
            ALLOCATE(recv%buffer_r3(recv%size,dim3))
           
            ireq=ireq+1
            CALL MPI_IRECV(recv%buffer_r3,recv%size*dim3,MPI_REAL8,recv%rank,recv%domain,comm_icosa, mpi_req(ireq),ierr)
          ENDDO
        
        ENDDO
        
        CALL MPI_WAITALL(nreq,mpi_req,status,ierr)
        
        DO ind=1,ndomain
          rval3d=>field(ind)%rval3d
        
          req=>request(ind)
          DO isend=1,req%nsend
            send=>req%send(isend)
            DEALLOCATE(send%buffer_r3)
          ENDDO
        
          DO irecv=1,req%nrecv
            recv=>req%recv(irecv)
            buffer_r3=>recv%buffer_r3
            value=>recv%value
            sgn=>recv%sign
            DO n=1,recv%size
              rval3d(value(n),:)=buffer_r3(n,:)*sgn(n)  
            ENDDO        
            DEALLOCATE(recv%buffer_r3)
          ENDDO
        
        ENDDO

      ELSE  IF (field(1)%ndim==4) THEN
      
        nreq=sum(request(:)%nsend)+sum(request(:)%nrecv)
        ALLOCATE(mpi_req(nreq))
        ALLOCATE(status(MPI_STATUS_SIZE,nreq))
    
        ireq=0
        DO ind=1,ndomain
          dim3=size(field(ind)%rval4d,2)
          dim4=size(field(ind)%rval4d,3)
          rval4d=>field(ind)%rval4d
        
          req=>request(ind)
          DO isend=1,req%nsend
            send=>req%send(isend)

            ALLOCATE(send%buffer_r4(send%size,dim3,dim4))
            buffer_r4=>send%buffer_r4
            value=>send%value
            DO n=1,send%size
              buffer_r4(n,:,:)=rval4d(value(n),:,:)
            ENDDO

            ireq=ireq+1
            CALL MPI_ISEND(buffer_r4,send%size*dim3*dim4,MPI_REAL8,send%rank,ind,comm_icosa, mpi_req(ireq),ierr)
          ENDDO
        
          DO irecv=1,req%nrecv
            recv=>req%recv(irecv)
            ALLOCATE(recv%buffer_r4(recv%size,dim3,dim4))
           
            ireq=ireq+1
            CALL MPI_IRECV(recv%buffer_r4,recv%size*dim3*dim4,MPI_REAL8,recv%rank,recv%domain,comm_icosa, mpi_req(ireq),ierr)
          ENDDO
        
        ENDDO
        
        CALL MPI_WAITALL(nreq,mpi_req,status,ierr)
        
        DO ind=1,ndomain
          rval4d=>field(ind)%rval4d
        
          req=>request(ind)
          DO isend=1,req%nsend
            send=>req%send(isend)
            DEALLOCATE(send%buffer_r4)
          ENDDO
        
          DO irecv=1,req%nrecv
            recv=>req%recv(irecv)
            buffer_r4=>recv%buffer_r4
            value=>recv%value
            sgn=>recv%sign
            DO n=1,recv%size
              rval4d(value(n),:,:)=buffer_r4(n,:,:)*sgn(n) 
            ENDDO        
            DEALLOCATE(recv%buffer_r4)
          ENDDO
        
        ENDDO
      
      ENDIF      
      
    ENDIF

    CALL trace_end("transfert_mpi")
    
  END SUBROUTINE transfert_request_mpi
   
  SUBROUTINE transfert_request_seq(field,request)
  USE field_mod
  USE domain_mod
  IMPLICIT NONE
    TYPE(t_field),POINTER :: field(:)
    TYPE(t_request),POINTER :: request(:)
    REAL(rstd),POINTER :: rval2d(:) 
    REAL(rstd),POINTER :: rval3d(:,:) 
    REAL(rstd),POINTER :: rval4d(:,:,:) 
    INTEGER :: ind
    TYPE(t_request),POINTER :: req
    INTEGER :: n
    REAL(rstd) :: var1,var2
    
    DO ind=1,ndomain
      req=>request(ind)
      rval2d=>field(ind)%rval2d
      rval3d=>field(ind)%rval3d
      rval4d=>field(ind)%rval4d
      
      IF (field(ind)%data_type==type_real) THEN
        IF (field(ind)%ndim==2) THEN
          DO n=1,req%size
            rval2d(req%target_ind(n))=field(req%src_domain(n))%rval2d(req%src_ind(n))*req%target_sign(n)
          ENDDO
        ELSE IF (field(ind)%ndim==3) THEN
          DO n=1,req%size
            rval3d(req%target_ind(n),:)=field(req%src_domain(n))%rval3d(req%src_ind(n),:)*req%target_sign(n)
          ENDDO
        ELSE IF (field(ind)%ndim==4) THEN
          DO n=1,req%size
            rval4d(req%target_ind(n),:,:)=field(req%src_domain(n))%rval4d(req%src_ind(n),:,:)*req%target_sign(n)
          ENDDO
        ENDIF
      ENDIF        

    ENDDO
    
  END SUBROUTINE transfert_request_seq
  
  
  SUBROUTINE gather_field(field_loc,field_glo)
  USE field_mod
  USE domain_mod
  USE mpi_mod
  USE mpipara
  IMPLICIT NONE
    TYPE(t_field),POINTER :: field_loc(:)
    TYPE(t_field),POINTER :: field_glo(:)
    INTEGER, ALLOCATABLE :: mpi_req(:)
    INTEGER, ALLOCATABLE :: status(:,:)
    INTEGER :: ireq,nreq
    INTEGER :: ind_glo,ind_loc    
  
    IF (.NOT. using_mpi) THEN
    
      DO ind_loc=1,ndomain
        IF (field_loc(ind_loc)%ndim==2) field_glo(ind_loc)%rval2d=field_loc(ind_loc)%rval2d
        IF (field_loc(ind_loc)%ndim==3) field_glo(ind_loc)%rval3d=field_loc(ind_loc)%rval3d
        IF (field_loc(ind_loc)%ndim==4) field_glo(ind_loc)%rval4d=field_loc(ind_loc)%rval4d
      ENDDO
    
    ELSE
          
      nreq=ndomain
      IF (mpi_rank==0) nreq=nreq+ndomain_glo 
      ALLOCATE(mpi_req(nreq))
      ALLOCATE(status(MPI_STATUS_SIZE,nreq))
    
    
      ireq=0
      IF (mpi_rank==0) THEN
        DO ind_glo=1,ndomain_glo
          ireq=ireq+1

          IF (field_glo(ind_glo)%ndim==2) THEN
            CALL MPI_IRECV(field_glo(ind_glo)%rval2d,size(field_glo(ind_glo)%rval2d) , MPI_REAL8 ,   &
                         domglo_rank(ind_glo),domglo_loc_ind(ind_glo), comm_icosa, mpi_req(ireq), ierr)
   
          ELSE IF (field_glo(ind_glo)%ndim==3) THEN
            CALL MPI_IRECV(field_glo(ind_glo)%rval3d,size(field_glo(ind_glo)%rval3d) , MPI_REAL8 ,   &
                         domglo_rank(ind_glo),domglo_loc_ind(ind_glo), comm_icosa, mpi_req(ireq), ierr)

          ELSE IF (field_glo(ind_glo)%ndim==4) THEN
            CALL MPI_IRECV(field_glo(ind_glo)%rval4d,size(field_glo(ind_glo)%rval4d) , MPI_REAL8 ,   &
                         domglo_rank(ind_glo),domglo_loc_ind(ind_glo), comm_icosa, mpi_req(ireq), ierr)
          ENDIF
         
        ENDDO
      ENDIF
  
      DO ind_loc=1,ndomain
        ireq=ireq+1

        IF (field_loc(ind_loc)%ndim==2) THEN
          CALL MPI_ISEND(field_loc(ind_loc)%rval2d,size(field_loc(ind_loc)%rval2d) , MPI_REAL8 ,   &
                         0, ind_loc, comm_icosa, mpi_req(ireq), ierr)
        ELSE IF (field_loc(ind_loc)%ndim==3) THEN
          CALL MPI_ISEND(field_loc(ind_loc)%rval3d,size(field_loc(ind_loc)%rval3d) , MPI_REAL8 ,   &
                         0, ind_loc, comm_icosa, mpi_req(ireq), ierr)
        ELSE IF (field_loc(ind_loc)%ndim==4) THEN
          CALL MPI_ISEND(field_loc(ind_loc)%rval4d,size(field_loc(ind_loc)%rval4d) , MPI_REAL8 ,   &
                         0, ind_loc, comm_icosa, mpi_req(ireq), ierr)
        ENDIF
      
      ENDDO
   
      CALL MPI_WAITALL(nreq,mpi_req,status,ierr)

    ENDIF
        
  END SUBROUTINE gather_field

   
  SUBROUTINE trace_in
  USE trace
  IMPLICIT NONE
  
    CALL trace_start("transfert_buffer")
  END SUBROUTINE trace_in              

  SUBROUTINE trace_out
  USE trace
  IMPLICIT NONE
  
    CALL trace_end("transfert_buffer")
  END SUBROUTINE trace_out              

END MODULE transfert_mpi_mod
      
        
        
        
      
